#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2026-02-09
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : xp

"""
Gets numpy or cupy depending on the value of cuda argument.
"""



# %% Libraries
import numpy as np
try:
    import cupy as cp
except ImportError:
    cp = None



# %% Function
def xp(cuda:bool) :
    '''
    Gets numpy or cupy depending on the value of cuda argument.
    
    Parameters
    ----------
    cuda : bool
        Tells to take numpy or cupy.

    Returns
    -------
    _xp : np or cp
        Library to use.

    Raises
    ------
    ImportError
        If cuda is asked but unavailable.

    Examples
    --------
    >>> from arrlp import xp
    ...
    >>> _xp = xp(cuda)
    '''

    if cuda and cp is None :
        raise ImportError('Cupy is not available for xp function')
    return cp if cuda else np



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2026-02-09
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : parallel_array

"""
Applies a function in parallel on stacks and channels dimensions.
"""



# %% Libraries
from joblib import Parallel, delayed
from numba import njit, prange
import numpy as np



# %% Function
def parallel_array(function, *args, outs, stacks, **kwargs) :
    '''
    Applies a function in parallel on stacks and channels dimensions.
    
    Parameters
    ----------
    function : function
        Function to apply in parallel.
    outs : tuple
        Tuple of out arrays on which to put results.
    *args : tuple
        Arrays on which to loop.
    stacks : bool
        True if looping on stacks.
    **kwargs : dict
        Fixed parameters to pass to function.

    Examples
    --------
    >>> from arrlp import parallel_array
    ...
    >>> parallel_array(xp.fft, arrays, outs=outs, stacks=True, kernel=kernel)
    '''

    if not isinstance(outs, tuple) :
        raise TypeError(f'outs should be tuple not {type(outs)}')
    outs = list(outs)

    # Infer dimensions
    shape = args[0].shape
    nstacks = shape[0]
    nchannels = shape[-1]

    # Loop on stacks
    if stacks :
        copy = Parallel(n_jobs=-1, backend="loky")(
            delayed(function)(
                *(arg[i] for arg in args), **kwargs
            )
            for i in range(nstacks)
        )

        results = (copy,) if len(outs) == 1 else zip(*copy)
        for i, (res, dst) in enumerate(zip(results, outs)):
            if dst is None :
                dst = np.empty_like(res[0], shape=(len(res), *res[0].shape))
                outs[i] = dst
            copystacks(list(res), dst)

    # Loop on channels
    else :
        copy = Parallel(n_jobs=-1, backend="loky")(
            delayed(function)(
                *(arg[..., i] for arg in args), **kwargs
            )
            for i in range(nchannels)
        )

        results = (copy,) if len(outs) == 1 else zip(*copy)
        for i, (res, dst) in enumerate(zip(results, outs)):
            if dst is None :
                dst = np.empty_like(res[0], shape=(*res[0].shape, len(res)))
                outs[i] = dst
            copychannels(list(res), dst)
    
    # End
    return outs[0] if len(outs) == 1 else tuple(outs)



@njit(parallel=True)
def copystacks(copyfrom, copyto):
    n = len(copyfrom)
    for i in prange(n):
        copyto[i] = copyfrom[i]

@njit(parallel=True)
def copychannels(copyfrom, copyto):
    n = len(copyfrom)
    for i in prange(n):
        copyto[..., i] = copyfrom[i]



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from arrlp import xp, parallel_array, img_fft
from arrlp import kernel as _kernel



# %% Function
def img_wiener(array, *,
        stacks=False, channels=False, parallel=False, cuda=False, print=None,
        kernel=None, power=2, balance=10**2, **kernel_kwargs) :
    '''
    Makes Wiener deconvolution.
    '''

    # Checks
    if parallel and not stacks and not channels :
        raise ValueError('Normal array (no stack of channel) cannot be calculated in parallel')
    # if parallel :
    #     import warnings
    #     warnings.warn('Parallel optimization is not effective in this function')
    # if cuda :
    #     import warnings
    #     warnings.warn('Cuda optimization is not effective in this function')
        

    # Init
    _xp = xp(cuda)
    array = _xp.asarray(array)
    out = _xp.empty_like(array)

    # Function info [update here]
    ndims = 2
    ins = (array,)
    outs = (out,)

    # Init
    nstacks, nchannels = array.shape[0], array.shape[-1]
    iterator = print.clock if print is not None else range

    # Wiener
    y, x = array.shape[int(stacks)], array.shape[int(stacks) + 1]
    if kernel is None : kernel = _kernel(ndims, shape=(y, x), cuda=cuda, **kernel_kwargs)
    fft_kernel = _xp.abs(img_fft(kernel, cuda=cuda))
    W = fft_kernel**(power-1) / (fft_kernel**power + fft_kernel.max() / balance)
    def func(array) :
        dtype = array.dtype
        fft = _xp.fft.fftshift(_xp.fft.fft2(array))
        return _xp.real(_xp.fft.ifft2(_xp.fft.ifftshift(fft * W))).astype(dtype)


    # Looping on axes
    if cuda or not parallel :

        match (stacks, channels) :
            case (False, False) :
                out[:] = func(array)
            case (True, False) :
                for i in iterator(nstacks) :
                    out[i] = func(array[i])
            case (False, True) :
                for j in iterator(nchannels) :
                    out[..., j] = func(array[..., j])
            case (True, True) :
                for i in iterator(nstacks) :
                    a, o = array[i], out[i]
                    for j in range(nchannels) :
                        o[..., j] = func(a[..., j])
        return out
    
    # Parallel
    if print is not None : print('Looping in parallel... [ETA not available]')

    match (stacks, channels) :

        case (True, False) :
            return parallel_array(func, *ins, outs=outs, stacks=True)

        case (False, True) :
            return parallel_array(func, *ins, outs=outs, stacks=False)

        case (True, True) :
            def function(array) :
                out = _xp.empty_like(array)
                for j in range(nchannels) :
                    out[..., j] = func(array[..., j])
                return out
            return parallel_array(function, *ins, outs=outs, stacks=True)
    




if __name__ == '__main__' :



    # Imports
    import numpy as np
    try :
        import cupy as cp
    except ImportError :
        cp = None
    from time import perf_counter



    # Parameter ~2**24 ; 2**8=256
    func = img_wiener
    nstacks = int(2**8)
    shape = (int(2**7), int(2**7))
    nchannels = int(2**2)



    # Timeit function
    def timeit(array, stacks, channels, parallel, cuda) :

        # Init
        _xp = xp(cuda)
        array = _xp.asarray(array)

        # Calculate
        print(f'\n** Testing stacks={stacks}, channels={channels}, parallel={parallel}, cuda={cuda} **')
        print('Compile run...')
        func(array, stacks=stacks, channels=channels, parallel=parallel, cuda=cuda, pixel=1, sigma=3)
        print('Run...')
        tic = perf_counter()
        out = func(array, stacks=stacks, channels=channels, parallel=parallel, cuda=cuda, pixel=1, sigma=3)
        toc = perf_counter()
        print(f'...took {(toc-tic)*1000:.3f}ms\n')
        if cuda : out = _xp.asnumpy(out)
        return out



    # Loops
    optimization = {'None': dict(cuda=False, parallel=False), 'Parallel': dict(cuda=False, parallel=True), 'Cuda': dict(cuda=True, parallel=False)}
    opti = ["None", "Parallel"] if cp is None else ["None", "Parallel", "Cuda"]
    for channels in [False, True] :
        for stacks in [False, True] :

            # Array
            match stacks, channels :
                case (True, True) :
                    _shape = (nstacks, *shape, nchannels)
                case(True, False) :
                    _shape = (nstacks, *shape)
                case(False, True) :
                    _shape = (*shape, nchannels)
                case(False, False) :
                    _shape = shape
            array = np.random.random(_shape)

            # Calculate
            results = {}
            for opt in opti:
                results[opt] = timeit(array, stacks, channels, **optimization[opt])
                if not channels and not stacks and opt == "None" :
                    break
            
            # Compare
            ref = results["None"]
            for opt, out in results.items():
                if opt == "None" : continue
                print(f"Checking correctness vs reference: {opt}")
                np.testing.assert_allclose(
                    ref, out,
                    rtol=1e-5,
                    atol=1e-6,
                    err_msg="Outputs differ between optimizations"
                )

                



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet



# %% Libraries
from arrlp import xp, axes, parallel_array



# %% Function
def temp(array, *,
        stacks=False, channels=False, parallel=False, cuda=False, print=None,
        **kwargs) :
    '''
    TODO.
    '''

    # Checks
    if parallel and not stacks and not channels :
        raise ValueError('Normal array (no stack of channel) cannot be calculated in parallel')
    # if parallel :
    #     import warnings
    #     warnings.warn('Parallel optimization is not effective in this function')
    # if cuda :
    #     import warnings
    #     warnings.warn('Cuda optimization is not effective in this function')
        

    # Init
    _xp = xp(cuda)
    array = _xp.asarray(array)
    # if out is None: out = _xp.empty_like(array)

    # Function info [update here]
    func = _xp.TODO
    ndims = 2
    ins = (array,)
    outs = (None,)

    # Init

    # Looping on axes
    if cuda or not parallel :
        _axes = axes(ndims, stacks)
        return func(array, axes=_axes, **kwargs) # TODO
    
    # Parallel

    match (stacks, channels) :

        case (True, False) :
            return parallel_array(func, *ins, outs=outs, stacks=True, **kwargs)

        case (False, True) :
            return parallel_array(func, *ins, outs=outs, stacks=False, **kwargs)

        case (True, True) :
            _axes = axes(ndims, False)
            return parallel_array(func, *ins, outs=outs, stacks=True, axes=_axes, **kwargs)
    




if __name__ == '__main__' :



    # Imports
    import numpy as np
    try :
        import cupy as cp
    except ImportError :
        cp = None
    from time import perf_counter



    # Parameter ~2**24 ; 2**8=256
    func = TODO
    nstacks = int(2**8)
    shape = (int(2**7), int(2**7))
    nchannels = int(2**2)



    # Timeit function
    def timeit(array, stacks, channels, parallel, cuda) :

        # Init
        _xp = xp(cuda)
        array = _xp.asarray(array)

        # Calculate
        print(f'\n** Testing stacks={stacks}, channels={channels}, parallel={parallel}, cuda={cuda} **')
        print('Compile run...')
        func(array, stacks=stacks, channels=channels, parallel=parallel, cuda=cuda)
        print('Run...')
        tic = perf_counter()
        out = func(array, stacks=stacks, channels=channels, parallel=parallel, cuda=cuda)
        toc = perf_counter()
        print(f'...took {(toc-tic)*1000:.3f}ms\n')
        if cuda : out = _xp.asnumpy(out)
        return out



    # Loops
    optimization = {'None': dict(cuda=False, parallel=False), 'Parallel': dict(cuda=False, parallel=True), 'Cuda': dict(cuda=True, parallel=False)}
    opti = ["None", "Parallel"] if cp is None else ["None", "Parallel", "Cuda"]
    for channels in [False, True] :
        for stacks in [False, True] :

            # Array
            match stacks, channels :
                case (True, True) :
                    _shape = (nstacks, *shape, nchannels)
                case(True, False) :
                    _shape = (nstacks, *shape)
                case(False, True) :
                    _shape = (*shape, nchannels)
                case(False, False) :
                    _shape = shape
            array = np.random.random(_shape)

            # Calculate
            results = {}
            for opt in opti:
                results[opt] = timeit(array, stacks, channels, **optimization[opt])
                if not channels and not stacks and opt == "None" :
                    break
            
            # Compare
            ref = results["None"]
            for opt, out in results.items():
                if opt == "None" : continue
                print(f"Checking correctness vs reference: {opt}")
                np.testing.assert_allclose(
                    ref, out,
                    rtol=1e-5,
                    atol=1e-7,
                    err_msg="Outputs differ between optimizations"
                )

                



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2025-08-28
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP

"""
A library that provides tools for arrays on CPU & GPU [any shape].
"""



# %% Source import
sources = {
'axes': 'arrlp.modules.axes_LP.axes',
'compress': 'arrlp.modules.compress_LP.compress',
'coordinates': 'arrlp.modules.coordinates_LP.coordinates',
'image': 'arrlp.modules.image_LP.image',
'kernel': 'arrlp.modules.kernel_LP.kernel',
'ndimage': 'arrlp.modules.ndimage_LP.ndimage',
'parallel_array': 'arrlp.modules.parallel_array_LP.parallel_array',
'signal': 'arrlp.modules.signal_LP.signal',
'volume': 'arrlp.modules.volume_LP.volume',
'xp': 'arrlp.modules.xp_LP.xp',
'img_wiener': 'arrlp.modules.image_LP._functions.img_wiener',
'img_correlate': 'arrlp.modules.image_LP._functions.img_correlate',
'img_fft': 'arrlp.modules.image_LP._functions.img_fft',
'img_ifft': 'arrlp.modules.image_LP._functions.img_ifft'
}



# %% Lazy imports
from corelp import getmodule
__getattr__, __all__ = getmodule(sources)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2026-02-10
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : ndimage

"""
Gets scipy.ndimage of cupyx.scipy.ndimage depending on the value of cuda argument.
"""



# %% Libraries
import scipy.ndimage as cpu_ndimage
try:
    import cupyx.scipy.ndimage as gpu_ndimage 
except ImportError:
    gpu_ndimage = None



# %% Function
def ndimage(cuda:bool) :
    '''
    Gets scipy.ndimage of cupyx.scipy.ndimage depending on the value of cuda argument.
    
    Parameters
    ----------
    cuda : bool
        Tells to take scipy or cupyx.scipy.

    Returns
    -------
    _ndimage : scipy.ndimage or cupyx.scipy.ndimage
        Library to use.

    Raises
    ------
    ImportError
        If cuda is asked but unavailable.

    Examples
    --------
    >>> from arrlp import ndimage
    ...
    >>> _ndimage = ndimage(cuda)
    '''

    if cuda and gpu_ndimage is None :
        raise ImportError('Cupy is not available for xp function')
    return gpu_ndimage if cuda else cpu_ndimage



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Date          : 2026-02-09
# Author        : Lancelot PINCET
# GitHub        : https://github.com/LancelotPincet
# Library       : arrLP
# Module        : axes

"""
Get axes on which to apply the xp function.
"""



# %% Libraries



# %% Function
def axes(ndim:int, stacks:bool) :
    '''
    Get axes on which to apply the xp function.
    
    Parameters
    ----------
    ndim : int
        Number of dimensions of array.

    Returns
    -------
    _axes : tuple
        axes on which to do calculation.

    Examples
    --------
    >>> from arrlp import axes
    ...
    >>> axes = axes(2, True)
    '''

    start = int(stacks)
    return tuple(range(start, start + ndim))



# %% Test function run
if __name__ == "__main__":
    from corelp import test
    test(__file__)